import pytest


@pytest.fixture
def record_payload():
    return {
        'domain': 'example.com',
        'record_name': '_acme-challenge.example.com',
        'record_content': '123456790',
        'record_ttl': 300
    }

@pytest.fixture
def txt_data_payload():
    return {"resource_records": [{"content": ['text'], "enabled": True}], "ttl": 300}



"""DNS Authenticator for G-Core."""

import logging
from typing import Any
from typing import Callable
from typing import Optional

from certbot import errors
from certbot.plugins import dns_common
from certbot.plugins.dns_common import CredentialsConfiguration

from . import api_gcore

logger = logging.getLogger(__name__)


class Authenticator(dns_common.DNSAuthenticator):
    """DNS Authenticator for G-Core

    This Authenticator uses the G-Core DNS API to fulfill a dns-01 challenge.
    """

    _docs_url = 'https://apidocs.gcorelabs.com/dns#section/Authentication'
    description = ('Obtain certificates using a DNS TXT record (if you are using G-Core for '
                   'DNS).')
    ttl = 300

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.token = self.email = self.password = None
        self.credentials: Optional[CredentialsConfiguration] = None

    @classmethod
    def add_parser_arguments(
            cls, add: Callable[..., None], default_propagation_seconds: int = 10
    ) -> None:
        super().add_parser_arguments(add, default_propagation_seconds)
        add('credentials', help='G-Core credentials INI file.')

    def more_info(self) -> str:
        return 'This plugin configures a DNS TXT record to respond to a dns-01 challenge using ' + \
               'the G-Core API.'

    def _validate_credentials(self, credentials: CredentialsConfiguration) -> None:
        self.token = credentials.conf('apitoken')
        self.email = credentials.conf('email')
        self.password = credentials.conf('password')
        if self.token:
            if self.email or self.password:
                raise errors.PluginError('{}: dns_gcore_email and dns_gcore_password are '
                                         'not needed when using an API Token'
                                         .format(credentials.confobj.filename))
        elif self.email or self.password:
            if not self.email:
                raise errors.PluginError('{}: dns_gcore_email is required when using a Global '
                                         'API Key. (should be email address associated with '
                                         'G-Core account)'.format(credentials.confobj.filename))
            if not self.password:
                raise errors.PluginError('{}: dns_gcore_password is required when using a '
                                         'Global API Key. (see {})'
                                         .format(credentials.confobj.filename, self._docs_url))
        else:
            raise errors.PluginError(
                '{}: Either dns_gcore_apitoken (recommended), or '
                'dns_gcore_email and dns_gcore_password are required.'
                ' (see {})'.format(credentials.confobj.filename, self._docs_url)
            )

    def _setup_credentials(self) -> None:
        self.credentials = self._configure_credentials(
            'credentials',
            'G-Core credentials INI file',
            None,
            self._validate_credentials
        )

    def _perform(self, domain: str, validation_name: str, validation: str) -> None:
        self._get_client().add_txt_record(domain, validation_name, validation, self.ttl)

    def _cleanup(self, domain: str, validation_name: str, validation: str) -> None:
        self._get_client().del_txt_record(domain, validation_name)

    def _get_client(self) -> "_GCoreClient":
        if not self.credentials:  # pragma: no cover
            raise errors.Error("Plugin has not been prepared.")
        if self.token:
            return _GCoreClient(token=self.token)
        return _GCoreClient(login=self.email, password=self.password)


class _GCoreClient:
    """
    G-Core client.
    """

    record_type = 'TXT'

    def __init__(self, *args, **kwargs) -> None:
        self.gcore = api_gcore.GCoreClient(*args, **kwargs)

    def add_txt_record(
            self, domain: str, record_name: str, record_content: str, record_ttl: int
    ) -> None:
        """
        Add a TXT record using the supplied information.

        :param str domain: The domain to use for verification.
        :param str record_name: The record name (typically beginning with '_acme-challenge.').
        :param str record_content: The record content (typically the challenge validation).
        :param int record_ttl: The record TTL (number of seconds that the record may be cached).
        :raises certbot.errors.PluginError: if an error occurs communicating with the G-Core DNS API
        """
        self.gcore.record_create(
            domain,
            record_name,
            self.record_type,
            data=self._data_for_txt(record_ttl, record_content),
        )
        logger.debug('Successfully added TXT record with record_name: %s', record_name)

    def del_txt_record(self, domain: str, record_name: str) -> None:
        """
        Delete a TXT record using the supplied information.

        :param str domain: The domain to use for verification.
        :param str record_name: The record name (typically beginning with '_acme-challenge.').
        :param str record_content: The record content (typically the challenge validation).
        """
        self.gcore.record_delete(domain, record_name, self.record_type)
        logger.debug('Successfully deleted TXT record.')

    @classmethod
    def _data_for_txt(cls, _ttl, _content):
        """Preparing data for TXT record."""
        return {'resource_records': [{'content': [_content], 'enabled': True}], 'ttl': _ttl}



import pytest
import responses

from certbot_dns_gcore.api_gcore import GCoreClient
from certbot_dns_gcore.dns_gcore import _GCoreClient


@pytest.mark.parametrize('kwargs', ({'token': None}, {'login': 'user'}, {'password': 'test'}))
def test_gcoreclient_fail(kwargs):
    # CHECK
    with pytest.raises(ValueError):
        _GCoreClient(**kwargs)


@responses.activate
@pytest.mark.parametrize('kwargs', ({'token': '123'}, {'login': 'user', 'password': 'test'}))
def test_add_txt_record_success(kwargs, record_payload):
    # INIT
    responses.add(
        responses.POST,
        f'{GCoreClient._auth_url}auth/jwt/login',
        json={'access': 'token'},
        status=200
    )
    responses.add(
        responses.POST,
        f'{GCoreClient._dns_api_url}zones/{record_payload["domain"]}/{record_payload["record_name"]}/TXT',
        json={}, status=200
    )

    # ACT # CHECK
    assert _GCoreClient(**kwargs).add_txt_record(**record_payload) is None


@responses.activate
@pytest.mark.parametrize('kwargs', ({'token': '123'}, {'login': 'user', 'password': 'test'}))
def test_del_txt_record_success(kwargs, record_payload):
    # INIT
    record_payload.pop('record_content')
    record_payload.pop('record_ttl')
    responses.add(
        responses.POST,
        f'{GCoreClient._auth_url}auth/jwt/login',
        json={'access': 'token'},
        status=200
    )
    responses.add(
        responses.DELETE,
        f'{GCoreClient._dns_api_url}zones/{record_payload["domain"]}/{record_payload["record_name"]}/TXT',
        json={}, status=200
    )

    # ACT # CHECK
    assert _GCoreClient(**kwargs).del_txt_record(**record_payload) is None


def test_data_for_txt(txt_data_payload):
    assert _GCoreClient._data_for_txt(300, 'text') == txt_data_payload



"""Wrapper for G-Core DNS API."""

import http
import logging
import urllib.parse

from requests import Session

logger = logging.getLogger(__name__)


class GCoreClientException(Exception):
    """G-Core DNS API client exception."""


class GCoreClient:
    """G-Core DNS API client."""

    _root_zones = 'zones'
    _dns_api_url = 'https://api.gcorelabs.com/dns/v2/'
    _auth_url = 'https://api.gcdn.co/'

    def __init__(self, token=None, login=None, password=None):
        self._session = Session()
        if token is not None:
            self._session.headers.update({'Authorization': f'APIKey {token}'})
        elif login is not None and password is not None:
            token = self._auth(self._auth_url, login, password)
            self._session.headers.update({'Authorization': f'Bearer {token}'})
        else:
            raise ValueError('either token or login & password must be set')

    def _auth(self, url, login, password):
        """Get auth token."""
        responce = self._session.request(
            'POST',
            self._build_url(url, 'auth', 'jwt', 'login'),
            json={'username': login, 'password': password},
        )
        responce.raise_for_status()
        return responce.json()['access']

    def _request(self, method, url, params=None, data=None):
        """Requests handler."""
        responce = self._session.request(method, url, params=params, json=data, timeout=30.0)
        if responce.status_code in (
            http.HTTPStatus.BAD_REQUEST,
            http.HTTPStatus.NOT_FOUND,
            http.HTTPStatus.INTERNAL_SERVER_ERROR,
            http.HTTPStatus.CONFLICT,
        ):
            logger.error(
                'Something went wrong. %r has been sent to %r: %s', data, url, responce.text,
            )
            raise GCoreClientException(responce.text)
        responce.raise_for_status()
        return responce

    def zone(self, zone_name):
        """Get DNS record."""
        return self._request(
            'GET', self._build_url(self._dns_api_url, self._root_zones, zone_name)
        ).json()

    def zone_create(self, zone_name):
        """Create DNS zone."""
        return self._request(
            'POST',
            self._build_url(self._dns_api_url, self._root_zones),
            data={'name': zone_name},
        ).json()

    def zone_records(self, zone_name):
        """List DNS records from zone."""
        url = self._build_url(self._dns_api_url, self._root_zones, zone_name, 'rrsets')
        rrsets = self._request('GET', url, params={'all': 'true'}).json()
        records = rrsets['rrsets']
        return records

    def record_create(self, zone_name, rrset_name, type_, data):
        """Create DNS record in zone."""
        self._request('POST', self._rrset_url(zone_name, rrset_name, type_), data=data)

    def record_update(self, zone_name, rrset_name, type_, data):
        """Update DNS record in zone."""
        self._request('PUT', self._rrset_url(zone_name, rrset_name, type_), data=data)

    def record_delete(self, zone_name, rrset_name, type_):
        """Delete DNS record in zome."""
        self._request('DELETE', self._rrset_url(zone_name, rrset_name, type_))

    def _rrset_url(self, zone_name, rrset_name, type_):
        """Build full DNS URL for request."""
        return self._build_url(self._dns_api_url, self._root_zones, zone_name, rrset_name, type_)

    @staticmethod
    def _build_url(base, *items):
        for i in items:
            base = base.strip('/') + '/'
            base = urllib.parse.urljoin(base, i)
        return base



from setuptools import find_packages
from setuptools import setup

version = '0.1.2'

install_requires = [
    'certbot>=1.23.0',
    'setuptools>=39.0.1',
]

docs_extras = [
    'Sphinx>=1.0',
    'sphinx_rtd_theme',
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='certbot-dns-gcore',
    version=version,
    description="G-Core DNS Authenticator plugin for Certbot",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/G-Core/gcore-dns-certbot-plugin',
    author="G-Core Labs",
    author_email='support@gcorelabs.com',
    license='Apache License 2.0',
    python_requires='>=3.8',
    packages=find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require={
        'docs': docs_extras,
    },
    entry_points={
        'certbot.plugins': [
            'dns-gcore = certbot_dns_gcore.dns_gcore:Authenticator',
        ],
    },
)
